# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2019
"""
Schemas for streams created with the :py:meth:`~streamsx.eventstreams.subscribe` method, and usable for 
streams terminated with the :py:meth:`~streamsx.eventstreams.publish`. All of these message types are keyed messages.
"""

from streamsx.topology.schema import StreamSchema
#
# Defines Message types with default attribute names and types.
_SPL_SCHEMA_STRING_MESSAGE = 'tuple<rstring message,rstring key>'
_SPL_SCHEMA_BLOB_MESSAGE = 'tuple<blob message,rstring key>'
_SPL_SCHEMA_STRING_MESSAGE_META = 'tuple<rstring message,rstring key,rstring topic,int32 partition,int64 offset,int64 messageTimestamp>'
_SPL_SCHEMA_BLOB_MESSAGE_META = 'tuple<blob message,rstring key,rstring topic,int32 partition,int64 offset,int64 messageTimestamp>'


class Schema:
    """
    Structured stream schemas for keyed messages for :py:meth:`~streamsx.eventstreams.subscribe`, 
    and for streams that are published by :py:meth:`~streamsx.eventstreams.publish` to an Event Streams topic.
    
    The schemas
    
    * :py:const:`StringMessage`
    * :py:const:`BinaryMessage`
    
    have the attributes ``message``, and ``key``. They vary in the type for the 
    ``message`` attribute and can be used for :py:meth:`~streamsx.eventstreams.subscribe` 
    and for the stream published with :py:meth:`~streamsx.eventstreams.publish`.
    
    The schemas
    
    * :py:const:`StringMessageMeta`
    * :py:const:`BinaryMessageMeta`
    
    have the attributes ``message``, ``key``, ``topic``, ``partition``, ``offset``, and ``messageTimestamp``. They vary in the type for the 
    ``message`` attribute and can be used for :py:meth:`~streamsx.eventstreams.subscribe` and :py:meth:`~streamsx.eventstreams.publish`.
    
    All schemas defined in this class are instances of `streamsx.topology.schema.StreamSchema`.
    
    The following sample uses structured schemas for publishing messages with keys to a 
    potentially partitioned topic in Event Streams. Then, it creates a consumer group 
    that subscribes to the topic, and processes the received messages in parallel channels
    partitioned by the message key::
    
        from streamsx.topology.topology import Topology
        from streamsx.topology.context import submit, ContextTypes
        from streamsx.topology.topology import Routing
        from streamsx.topology.schema import StreamSchema
        from streamsx.eventstreams.schema import Schema
        import streamsx.eventstreams as evst
        
        import random
        import time
        import json
        from datetime import datetime
        
        
        # Define a callable source for data that we push into Event Streams
        class SensorReadingsSource(object):
            def __call__(self):
                # This is just an example of using generated data,
                # Here you could connect to db, generate data,
                # connect to data set, open file, ...
                i = 0
                # wait that the consumer is ready before we start creating data
                time.sleep(20.0)
                while(i < 10000):
                    time.sleep(0.01)   # 100 per second 
                    i = i + 1
                    sensor_id = random.randint(1, 100)
                    reading = {}
                    reading["sensor_id"] = "sensor_" + str(sensor_id)
                    reading["value"] = random.random() * 3000
                    reading["ts"] = int(datetime.now().timestamp())
                    yield reading
        
        
        # parses the JSON in the message and adds the attributes to a tuple
        def flat_message_json(tuple):
            messageAsDict = json.loads(tuple['message'])
            tuple.update(messageAsDict)
            return tuple
        
        
        # calculate a hash code of a string in a consistent way
        # needed for partitioned parallel streams
        def string_hashcode(s):
            h = 0
            for c in s:
                h = (31 * h + ord(c)) & 0xFFFFFFFF
            return ((h + 0x80000000) & 0xFFFFFFFF) - 0x80000000
        
        topology = Topology('EventStreamsParallel')
        
        #
        # the producer part
        #
        # create the data and map them to the attributes 'message' and 'key' of the
        # 'Schema.StringMessage' schema for Kafka, so that we have messages with keys
        sensorStream = topology.source(
            SensorReadingsSource(),
            "RawDataSource"
            ).map(
                func=lambda reading: {'message': json.dumps(reading),
                                      'key': reading['sensor_id']},
                name="ToKeyedMessage",
                schema=Schema.StringMessage)
        # assume, we have created an application configuration with name 'messagehub'
        eventStreamsSink = evst.publish(
            sensorStream,
            topic="threePartitionTopic1",
            credentials='messagehub',
            name="SensorPublish")
        
        
        #
        # the consumer side
        #
        # subscribe, create a consumer group with 3 consumers
        consumerSchema = Schema.StringMessageMeta
        received = evst.subscribe(
            topology,
            topic="threePartitionTopic1",
            schema=consumerSchema,
            group='my_consumer_group',
            credentials='messagehub',
            name="SensorSubscribe"
            ).set_parallel(3).end_parallel()
        
        # start a different parallel region partitioned by message key,
        # so that each key always goes into the same parallel channel
        receivedParallelPartitioned = received.parallel(
            5,
            routing=Routing.HASH_PARTITIONED,
            func=lambda x: string_hashcode(x['key']))
        
        # schema extension, here we use the Python 2.7, 3 way
        flattenedSchema = consumerSchema.extend(
            StreamSchema('tuple<rstring sensor_id, float64 value, int64 ts>'))
        
        receivedParallelPartitionedFlattened = receivedParallelPartitioned.map(
            func=flat_message_json,
            name='JSON2Attributes',
            schema=flattenedSchema)
        
        # validate by remove negativ and zero values from the streams,
        # pass only positive vaues and timestamps
        receivedValidated = receivedParallelPartitionedFlattened.filter(
            lambda tup: (tup['value'] > 0) and (tup['ts'] > 0),
            name='Validate')
        
        # end parallel processing and print the combined streams to stdout log
        receivedValidated.end_parallel().print()
        
        submit(ContextTypes.STREAMING_ANALYTICS_SERVICE, topology)
    
    """

    StringMessage = StreamSchema (_SPL_SCHEMA_STRING_MESSAGE)
    """
    Stream schema with message and key, both being strings.

    The schema defines following attributes
    
    * message(str) - the message content
    * key(str) - the key for partitioning
    
    This schema can be used for both :py:meth:`~streamsx.eventstreams.subscribe`, 
    and for streams that are published by :py:meth:`~streamsx.eventstreams.publish`.

     .. versionadded:: 1.2
    """

    StringMessageMeta = StreamSchema (_SPL_SCHEMA_STRING_MESSAGE_META)
    """
    Stream schema with message, key, and message meta data, where both message and key are strings.
    This schema can be used for :py:meth:`~streamsx.eventstreams.subscribe`.
    
    The schema defines following attributes
    
    * message(str) - the message content
    * key(str) - the key for partitioning
    * topic(str) - the Event Streams topic
    * partition(int) - the topic partition number (32 bit)
    * offset(int) - the offset of the message within the topic partition (64 bit)
    * messageTimestamp(int) - the message timestamp in milliseconds since epoch (64 bit)

     .. versionadded:: 1.2
    """

    BinaryMessage = StreamSchema (_SPL_SCHEMA_BLOB_MESSAGE)
    """
    Stream schema with message and key, where the message is a binary object (sequence of bytes), and the key is a string.

    The schema defines following attributes
    
    * message(bytes) - the message content
    * key(str) - the key for partitioning
    
    This schema can be used for both :py:meth:`~streamsx.eventstreams.subscribe`, 
    and for streams that are published by :py:meth:`~streamsx.eventstreams.publish`.

     .. versionadded:: 1.2
    """

    BinaryMessageMeta = StreamSchema (_SPL_SCHEMA_BLOB_MESSAGE_META)
    """
    Stream schema with message, key, and message meta data, where the message is a binary object (sequence of bytes), and the key is a string.
    This schema can be used for :py:meth:`~streamsx.eventstreams.subscribe`.
    
    The schema defines following attributes
    
    * message(bytes) - the message content
    * key(str) - the key for partitioning
    * topic(str) - the Event Streams topic
    * partition(int) - the topic partition number (32 bit)
    * offset(int) - the offset of the message within the topic partition (64 bit)
    * messageTimestamp(int) - the message timestamp in milliseconds since epoch (64 bit)

     .. versionadded:: 1.2
    """

    pass



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017

"""
Overview
++++++++

`IBM® Event Streams <https://www.ibm.com/cloud/event-streams>`_ is a fully managed, cloud-based messaging service. It is built on Apache Kafka and is available through IBM Cloud® as a Service.

This module allows a Streams application to :py:func:`subscribe <subscribe>` a
message queue as a stream and :py:func:`publish <publish>` messages on a queue from a stream
of tuples.

Credentials
+++++++++++

Event Streams credentials are defined using a Streams application configuration 
or setting the Event Streams service credentials JSON directly to the 
``credentials`` parameter of the functions as ``dict`` type.

Credentials Sample::

    import json
    import streamsx.eventstreams as eventstreams
    from streamsx.topology.schema import CommonSchema
    from streamsx.topology.topology import Topology
    
    topology = Topology ('CredentialsExample')
    
    myCredentialsJsonString = \"\"\"
    {
        your_credentials_json_formatted
        in multiple lines
    }
    \"\"\"
    myCredentials = json.loads(myCredentialsJsonString)
    stream = eventstreams.subscribe(topology, topic="topic",
        schema=CommonSchema.String, credentials=myCredentials)


By default an application configuration named `messagehub` is used,
a different configuration can be specified using the ``credentials``
parameter to :py:func:`subscribe` or :py:func:`publish`.

The application configuration must contain the property ``messagehub.creds`` with a value of the raw Event Streams service credentials JSON.

Messages
++++++++

The schema of the stream defines how messages are handled.

* ``CommonSchema.String`` - Each message is a UTF-8 encoded string.
* ``CommonSchema.Json`` - Each message is a UTF-8 encoded serialized JSON object.
* :py:const:`~schema.Schema.StringMessage` - structured schema with message and key
* :py:const:`~schema.Schema.BinaryMessage` - structured schema with message and key
* :py:const:`~schema.Schema.StringMessageMeta` - structured schema with message, key, and message meta data
* :py:const:`~schema.Schema.BinaryMessageMeta` - structured schema with message, key, and message meta data

No other formats are supported.

Sample
++++++

A simple hello world example of a Streams application publishing to
a topic and the same application consuming the same topic::

    from streamsx.topology.topology import Topology
    from streamsx.topology.schema import CommonSchema
    from streamsx.topology.context import submit, ContextTypes
    import streamsx.eventstreams as eventstreams
    import time

    def delay (v):
        time.sleep (5.0)
        return True

    topology = Topology ('EventStreamsHelloWorld')

    to_evstr = topology.source (['Hello', 'World!'])
    to_evstr = to_evstr.as_string()
    # delay tuple by tuple
    to_evstr = to_evstr.filter (delay)

    # Publish a stream to Event Streams using HELLO topic
    eventstreams.publish (to_evstr, topic='HELLO')

    # Subscribe to same topic as a stream
    from_evstr = eventstreams.subscribe (topology, schema=CommonSchema.String, topic='HELLO')

    # You'll find the Hello World! in stdout log file:
    from_evstr.print()

    submit (ContextTypes.STREAMING_ANALYTICS_SERVICE, topology)

"""

__version__ = '1.3.0'

__all__ = [
    'download_toolkit',
    'configure_connection',
    'subscribe',
    'publish'
    ]
from streamsx.eventstreams._eventstreams import subscribe, publish, configure_connection, download_toolkit



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017,2018

from tempfile import gettempdir
import json
import streamsx.spl.op
import streamsx.spl.types
from streamsx.topology.schema import CommonSchema
from streamsx.eventstreams.schema import Schema
from streamsx.toolkits import download_toolkit

_TOOLKIT_NAME = 'com.ibm.streamsx.messagehub'


def _add_toolkit_dependency(topo):
    # IMPORTANT: Dependency of this python wrapper to a specific toolkit version
    # This is important when toolkit is not set with streamsx.spl.toolkit.add_toolkit (selecting toolkit from remote build service)
    # messagehub toolkit >= 1.7.0 support the 'credentials' parameter were we can pass JSON directly to the operators
    streamsx.spl.toolkit.add_toolkit_dependency(topo, _TOOLKIT_NAME, '[1.7.0, 3.0.0)')


def _add_credentials_file(topology, credentials):
    """
    Adds a file dependency to the topology.
    The file contains the credentials as JSON.
    The filename in the bundle is ``etc/eventstreams.json``.
    """
    if credentials is None:
        raise TypeError(credentials)
    file_name = 'eventstreams.json'
    tmpdirname = gettempdir()
    tmpfile = tmpdirname + '/' + file_name
    with open(tmpfile, "w") as json_file:
        json_file.write(json.dumps(credentials))

    topology.add_file_dependency(tmpfile, 'etc')
    fName = 'etc/'+file_name
    print("Adding file dependency " + fName + " to the topology " + topology.name)
    return fName


def download_toolkit(url=None, target_dir=None):
    r"""Downloads the latest streamsx.messagehub toolkit from GitHub.

    Example for updating the toolkit for your topology with the latest toolkit from GitHub::

        import streamsx.eventstreams as es
        # download the toolkit from GitHub
        toolkit_location = es.download_toolkit()
        # add the toolkit to topology
        streamsx.spl.toolkit.add_toolkit(topology, toolkit_location)

    Example for updating the topology with a specific version of the streamsx.messagehub toolkit using an URL::

        import streamsx.eventstreams as es
        url202 = 'https://github.com/IBMStreams/streamsx.messagehub/releases/download/v2.0.2/com.ibm.streamsx.messagehub-2.0.2.tgz'
        toolkit_location = es.download_toolkit(url=url202)
        streamsx.spl.toolkit.add_toolkit(topology, toolkit_location)

    Args:
        url(str): Link to toolkit archive (\*.tgz) to be downloaded. Use this parameter to 
            download a specific version of the toolkit.
        target_dir(str): the directory where the toolkit is unpacked to. If a relative path is given,
            the path is appended to the system temporary directory, for example to /tmp on Unix/Linux systems.
            If target_dir is ``None`` a location relative to the system temporary directory is chosen.

    Returns:
        str: the location of the downloaded streamsx.messagehub toolkit

    .. note:: This function requires an outgoing Internet connection
    .. versionadded:: 1.3
    """
    _toolkit_location = streamsx.toolkits.download_toolkit (toolkit_name=_TOOLKIT_NAME, url=url, target_dir=target_dir)
    return _toolkit_location


def configure_connection(instance, name='messagehub', credentials=None):
    """Configures IBM Streams for a certain connection.


    Creates an application configuration object containing the required properties with connection information.


    Example for creating a configuration for a Streams instance with connection details::


        streamsx.rest import Instance
        import streamsx.topology.context
        from icpd_core import icpd_util

        cfg = icpd_util.get_service_instance_details(name='your-streams-instance')
        cfg[streamsx.topology.context.ConfigParams.SSL_VERIFY] = False
        instance = Instance.of_service(cfg)
        app_cfg = configure_connection(instance, credentials='my_crdentials_json')


    Args:
        instance(streamsx.rest_primitives.Instance): IBM Streams instance object.
        name(str): Name of the application configuration, default name is 'messagehub'.
        credentials(str|dict): The service credentials for Eventstreams.
    Returns:
        Name of the application configuration.

    .. warning:: The function can be used only in IBM Cloud Pak for Data.
    .. versionadded:: 1.1
    """

    description = 'Eventstreams credentials'
    properties = {}
    if credentials is None:
        raise TypeError(credentials)

    if isinstance(credentials, dict):
        properties['messagehub.creds'] = json.dumps(credentials)
    else:
        properties['messagehub.creds'] = credentials

    # check if application configuration exists
    app_config = instance.get_application_configurations(name=name)
    if app_config:
        print('update application configuration: ' + name)
        app_config[0].update(properties)
    else:
        print('create application configuration: ' + name)
        instance.create_application_configuration(name, properties, description)
    return name


def subscribe(topology, topic, schema, group=None, credentials=None, name=None):
    """Subscribe to messages from Event Streams (Message Hub) for a topic.

    Adds an Event Streams consumer that subscribes to a topic
    and converts each consumed message to a stream tuple.

    Args:
        topology(Topology): Topology that will contain the stream of messages.
        topic(str): Topic to subscribe messages from.
        schema(StreamSchema): Schema for returned stream.
        group(str): Kafka consumer group identifier. When not specified it default to the job name with `topic` appended separated by an underscore.
        credentials(dict|str): Credentials in JSON or name of the application configuration containing the credentials for the Event Streams service. When set to ``None`` the application configuration ``messagehub`` is used.
        name(str): Consumer name in the Streams context, defaults to a generated name.

    Returns:
         Stream: Stream containing messages.
    """
    if topic is None:
        raise TypeError(topic)
    msg_attr_name = None
    if schema is CommonSchema.Json:
        msg_attr_name = 'jsonString'
    elif schema is CommonSchema.String:
        msg_attr_name = 'string'
    elif schema is Schema.BinaryMessage:
        # msg_attr_name = 'message'
        pass
    elif schema is Schema.StringMessage:
        # msg_attr_name = 'message'
        pass
    elif schema is Schema.BinaryMessageMeta:
        # msg_attr_name = 'message'
        pass
    elif schema is Schema.StringMessageMeta:
        # msg_attr_name = 'message'
        pass
    else:
        raise ValueError(schema)

    if group is None:
        group = streamsx.spl.op.Expression.expression('getJobName() + "_" + "' + str(topic) + '"')

    if name is None:
        name = topic

    # check if it's the credentials for the service
    if isinstance(credentials, dict):
        appConfigName = None
    else:
        appConfigName = credentials

    _op = _MessageHubConsumer(topology, schema=schema, outputMessageAttributeName=msg_attr_name, appConfigName=appConfigName, topic=topic, groupId=group, name=name)
    if (appConfigName is None) and (credentials is not None):
        _op.params['messageHubCredentialsFile'] = _add_credentials_file(topology, credentials)

    return _op.stream


def publish(stream, topic, credentials=None, name=None):
    """Publish Event Streams messages to a topic.

    Adds an Event Streams producer where each tuple on `stream` is
    published as a message into IBM Event Streams cloud service.

    Args:
        stream(Stream): Stream of tuples to published as messages.
        topic(str): Topic to publish messages to.
        credentials(dict|str): Credentials in JSON or name of the application configuration containing the credentials for the Event Streams service. When set to ``None`` the application configuration ``messagehub`` is used.
        name(str): Producer name in the Streams context, defaults to a generated name.

    Returns:
        streamsx.topology.topology.Sink: Stream termination.
    """
    if topic is None:
        raise TypeError(topic)
    msg_attr_name = None
    streamSchema = stream.oport.schema
    if streamSchema == CommonSchema.Json:
        msg_attr_name = 'jsonString'
    elif streamSchema == CommonSchema.String:
        msg_attr_name = 'string'
    elif streamSchema is Schema.BinaryMessage:
        # msg_attr_name = 'message'
        pass
    elif streamSchema is Schema.StringMessage:
        # msg_attr_name = 'message'
        pass
    else:
        raise ValueError(streamSchema)

    # check if it's the credentials for the service
    if isinstance(credentials, dict):
        appConfigName = None
    else:
        appConfigName = credentials

    _op = _MessageHubProducer(stream, appConfigName=appConfigName, topic=topic, name=name)
    if (appConfigName is None) and (credentials is not None):
        _op.params['messageHubCredentialsFile'] = _add_credentials_file(stream.topology, credentials)

    # create the input attribute expressions after operator _op initialization
    if msg_attr_name is not None:
        _op.params['messageAttribute'] = _op.attribute(stream, msg_attr_name)
#    if keyAttributeName is not None:
#        params['keyAttribute'] = _op.attribute(stream, keyAttributeName)
#    if partitionAttributeName is not None:
#        params['partitionAttribute'] = _op.attribute(stream, partitionAttributeName)
#    if timestampAttributeName is not None:
#        params['timestampAttribute'] = _op.attribute(stream, timestampAttributeName)
#    if topicAttributeName is not None:
#        params['topicAttribute'] = _op.attribute(stream, topicAttributeName)

    return streamsx.topology.topology.Sink(_op)


class _MessageHubConsumer(streamsx.spl.op.Source):
    def __init__(self, topology, schema,
                 vmArg=None,
                 appConfigName=None,
                 clientId=None,
                 messageHubCredentialsFile=None,
                 outputKeyAttributeName=None,
                 outputMessageAttributeName=None,
                 outputTimestampAttributeName=None,
                 outputOffsetAttributeName=None,
                 outputPartitionAttributeName=None,
                 outputTopicAttributeName=None,
                 partition=None,
                 propertiesFile=None,
                 startPosition=None,
                 startTime=None,topic=None,
                 triggerCount=None,
                 userLib=None,
                 groupId=None,
                 name=None):
        kind = "com.ibm.streamsx.messagehub::MessageHubConsumer"
        # inputs = None
        schemas = schema
        params = dict()
        if vmArg is not None:
            params['vmArg'] = vmArg
        if appConfigName is not None:
            params['appConfigName'] = appConfigName
        if clientId is not None:
            params['clientId'] = clientId
        if messageHubCredentialsFile is not None:
            params['messageHubCredentialsFile'] = messageHubCredentialsFile
        if outputKeyAttributeName is not None:
            params['outputKeyAttributeName'] = outputKeyAttributeName
        if outputMessageAttributeName is not None:
            params['outputMessageAttributeName'] = outputMessageAttributeName
        if outputTimestampAttributeName is not None:
            params['outputTimestampAttributeName'] = outputTimestampAttributeName
        if outputOffsetAttributeName is not None:
            params['outputOffsetAttributeName'] = outputOffsetAttributeName
        if outputPartitionAttributeName is not None:
            params['outputPartitionAttributeName'] = outputPartitionAttributeName
        if outputTopicAttributeName is not None:
            params['outputTopicAttributeName'] = outputTopicAttributeName
        if partition is not None:
            params['partition'] = partition
        if propertiesFile is not None:
            params['propertiesFile'] = propertiesFile
        if startPosition is not None:
            params['startPosition'] = startPosition
        if startTime is not None:
            params['startTime'] = startTime
        if topic is not None:
            params['topic'] = topic
        if triggerCount is not None:
            params['triggerCount'] = triggerCount
        if userLib is not None:
            params['userLib'] = userLib
        if groupId is not None:
            params['groupId'] = groupId
        super(_MessageHubConsumer, self).__init__(topology, kind, schemas, params, name)


class _MessageHubProducer(streamsx.spl.op.Sink):
    def __init__(self, stream,
                 vmArg=None,
                 appConfigName=None,
                 messageHubCredentialsFile=None,
                 propertiesFile=None,
                 topic=None,
                 userLib=None,
                 name=None):
        # topology = stream.topology
        kind = "com.ibm.streamsx.messagehub::MessageHubProducer"
        params = dict()
        if vmArg is not None:
            params['vmArg'] = vmArg
        if appConfigName is not None:
            params['appConfigName'] = appConfigName
        if messageHubCredentialsFile is not None:
            params['messageHubCredentialsFile'] = messageHubCredentialsFile
        if propertiesFile is not None:
            params['propertiesFile'] = propertiesFile
        if topic is not None:
            params['topic'] = topic
        if userLib is not None:
            params['userLib'] = userLib
        super(_MessageHubProducer, self).__init__(kind, stream, params, name)
